import os
import subprocess
from pathlib import Path

def run_script_in_venv(venv_path, script_path, params=''):
    """在虚拟环境中运行脚本并传递参数"""
    python_exec = os.path.join(
        venv_path,
        'Scripts' if os.name == 'nt' else 'bin',
        'python.exe' if os.name == 'nt' else 'python'
    )
    if params == '':
        cmd = [python_exec, script_path]
    else:
        cmd = [python_exec, script_path, '-p', params]

    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        check=False
    )

    return {
        'success': result.returncode == 0,
        'returncode': result.returncode,
        'stdout': result.stdout,
        'stderr': result.stderr
    }

def get_files_in_dir(dir_path):
    """使用pathlib获取目录下一层的所有文件全路径"""
    path = Path(dir_path)
    return [str(file) for file in path.iterdir() if file.is_file()]

if __name__ == '__main__':
    venv_env = ''
    code_str = '<tool_call>Terminate</tool_call>'
    print(code_str.find('<tool_call>'))
    print(code_str.find('</tool_call>'))
    print(code_str.find('Terminate'))


FUNC_DESC = dict()
TAGS = ['资产', '主体', '产品']
VALID_FUNCS = ['percentile', 'sum', 'mean', 'count', 'single_ann_return', 'period_return', 'agg', '近N天', '近N月',
               '平台日', 'min', 'max', 'median', 'filter', 'sort', 'limit', 'alias', 'merge', 'select', '指标', 'concat'] + TAGS

FUNC_DESC['资产'] = '''
函数功能: 根据资产标签，筛选相应资产明细。并根据需要提取的字段，提取资产类指标。
参数列表:
    1. 资产标签列表([list]): 资产标签列表
    2. 资产指标列表([list]): 资产和持仓相关指标列表
    3. 穿透(string)：表示需要穿透至资管计划持仓, 该参数为可选参数，当不指定该参数时为不穿透情况
返回数据: 包含基本字段及所选字段的二维表格数据
示例程序:
    >>> 所有资产 = 资产([], ['全价市值', '剩余期限（天）'])
    >>> 回购持仓_明细 = 资产(['质押式回购', '买断式回购'], ['剩余期限（天）'], '穿透')
    >>> 逆回购持仓 = 资产(['逆回购'], ['全价市值'])
'''

FUNC_DESC['产品'] = '''
函数功能: 根据产品标签，筛选相应产品明细。并根据需要提取的字段，提取产品类指标。
参数列表:
    1. 产品标签列表([list]): 产品标签列表
    2. 产品指标列表([list]): 产品相关指标列表
返回数据: 包含基本字段及所选字段的二维表格数据
示例程序:
    >>> 所有产品 = 产品([], ['当日净资产', '上一日净资产'])
    >>> 公募产品 = 产品(['公募产品'], ['上一日净资产'])
'''

FUNC_DESC['指标'] = '''
函数功能: 根据金融指标库、指标库下相关指标列表和参数条件提取的指标数据。
参数列表:
    1. 金融指标库名称(string): 金融指标库名称
    2. 指标列表([list]): 所选金融指标库下的指标列表
    3. 参数列表([list]): 查询是需要代入的参数
返回数据: 包含基本字段及所选字段的二维表格数据
示例程序:
    >>> 产品指标 = 指标(['估值汇总指标库'], ['产品代码', '日期', '总资产', '净资产'], {'port_tags': ['公募产品'], 'start_date': '2024-11-24'})
    >>> 资产指标 = 指标(['估值明细指标库'], ['产品代码', '资产代码', '资产简称', '持仓市值'], {'port_tags': ['公募产品'], 'asset_tags': ['债券']})
'''

FUNC_DESC['agg'] = '''
函数功能: 将一个表格数据按照某几个字段进行分组后，将指定字段进行聚合计算。
参数列表:
    1. 表格变量: 需要分组聚合的目标表格变量
    2. 分组字段列表([list]): 分组字段列表
    3. 聚合计算方式({dict}): 需要进行聚合字段的计算方式，计算方式可以选择sum, count, mean, max, min
    4. 分组后重命名列表: 重命名列表 
返回数据: 分组聚合后的二维表格数据
示例程序:
    >>> 公募产品聚合 = agg(公募产品, ['产品代码', '产品简称'], {'当日净资产': 'sum', '上一日净资产': 'mean'})
    >>> 债券类资产聚合 = agg(债券明细, ['产品代码', '产品简称', '资产代码'], {'全价市值': 'sum', '当日净资产': 'first'}, {'全价市值': '聚合后市值'})
'''

FUNC_DESC['select'] = '''
函数功能: 从一个表格数据中提取相应字段数据。
参数列表:
    1. 表格变量: 目标表格变量
    2. 提取字段列表([list]): 提取字段列表 
返回数据: 提取后字段后的表格数据，注意：除了包含指定提取的字段外，也包括产品代码、产品名称、资产代码、资产名称、日期等基础字段
示例程序:
    >>> 选择后公募产品 = select(公募产品, ['当日净资产', '总资产'])
    >>> 选择后债券类资产 = select(债券类资产, ['估值类型', '全价市值'])
'''

FUNC_DESC['merge'] = '''
函数功能: 将两个表格数据进行关联处理
参数列表:
    1. 表格变量1: 第一个表格变量 
    2. 表格变量2: 第二个表格变量
    3. 关联列表([list]): 关联字段列表，注意列表中的字段必须同时存在于两个表格变量中
    4. 关联方式(string): left - 左关联; right - 右关联; inner - 内关联; outer - 全关联  
返回数据: 返回关联后的表格数据
示例程序:
    >>> 选择后公募产品 = merge(债券明细, 现金类产品, ['产品代码', '产品简称'], 'left')
'''

FUNC_DESC['concat'] = '''
函数功能: 将两个具有相同列的表格数据进行垂直连接
参数列表:
    1. 表格变量1: 第一个表格变量 
    2. 表格变量2: 第二个表格变量  
返回数据: 返回垂直连接后的表格数据
示例程序:
    >>> 全部库 = concat(债券库, 回购库)
'''

FUNC_DESC['filter'] = '''
函数功能: 根据相应的筛选条件，选择表格数据中符合筛选条件的行生成新的表格数据
参数列表:
    1. 表格变量: 目标表格数据变量 
    2. 筛选条件(string): 注意 筛选条件中指明的字段必须在目标表格变量中
    3. 关联列表([list]): 关联字段列表，注意列表中的字段必须同时存在于两个表格变量中
    4. 关联方式(string): left - 左关联; right - 右关联; inner - 内关联; outer - 全关联  
返回数据: 返回关联后的表格数据
示例程序:
    >>> 选择后债券明细 = filter(债券明细, '全价市值 > 3000000')
    >>> 选择后债券明细 = filter(债券明细, '(全价市值 > 3000000) & (全价市值 > 5000000)')
    >>> 选择后债券明细 = filter(债券明细, '(剩余期限 > 100) ｜ (信用债 == "是")')
'''

FUNC_DESC['alias'] = '''
函数功能: 将表格数据中的某一列进行重新命名
参数列表:
    1. 表格变量: 目标表格数据变量 
    2. 新字段名: 改变后的字段名称
    3. 旧字段名: 改变前的字段名称，注意：当不提供旧字段名，语言会根据表格中通过四则运算后的字段进行命名
返回数据: 返回关联后的表格数据
示例程序:
    >>> 命名后债券明细 = alias(债券明细, '市值', '全价市值')
    >>> 四则运算后明细 = alias(债券明细.数量 * 债券明细.价格, '计算后市值')
'''

FUNC_DESC['sort'] = '''
函数功能: 将一个表格按照其中一个字段进行升序或降序排列
参数列表:
    1. 表格变量: 表格变量数据 
    2. 需要排序的字段名称(string): 表格中需要按照其排序的字段名称
    3. 排序方式(string): -'升序', '降序' 
返回数据: 返回排序后的表格数据
示例程序:
    >>> 排序后指标库 = sort(资产指标库, ['持仓市值'], '降序')
'''

FUNC_DESC['max'] = '''
函数功能: 求一个表格中一个字段列的最大值
参数列表:
    1. 表格变量: 表格变量数据 
    2. 字段名称(string): 表格中的字段名称
返回数据: 返回所选列的最大值
示例程序:
    >>> 最大持仓市值 = max(资产指标库, '持仓市值')
'''

FUNC_DESC['min'] = '''
函数功能: 求一个表格中一个字段列的最小值
参数列表:
    1. 表格变量: 表格变量数据 
    2. 字段名称(string): 表格中的字段名称
返回数据: 返回所选列的最小值
示例程序:
    >>> 最小持仓市值 = min(资产指标库, '持仓市值')
'''

FUNC_DESC['mean'] = '''
函数功能: 求一个表格中一个字段列的平均值
参数列表:
    1. 表格变量: 表格变量数据 
    2. 字段名称(string): 表格中的字段名称
返回数据: 返回所选列的平均值
示例程序:
    >>> 平均持仓市值 = mean(资产指标库, '持仓市值')
'''

FUNC_DESC['median'] = '''
函数功能: 求一个表格中一个字段列的中位数
参数列表:
    1. 表格变量: 表格变量数据 
    2. 字段名称(string): 表格中的字段名称
返回数据: 返回所选列的平均值
示例程序:
    >>> 持仓中位数 = median(资产指标库, '持仓市值')
'''

FUNC_DESC['count'] = '''
函数功能: 求一个表格中一个字段列的个数
参数列表:
    1. 表格变量: 表格变量数据 
    2. 字段名称(string): 表格中的字段名称
返回数据: 返回所选列的个数
示例程序:
    >>> 持仓个数 = count(资产指标库, '持仓市值')
'''

FUNC_DESC['limit'] = '''
函数功能: 返回一个表格的前N行
参数列表:
    1. 表格变量: 表格变量数据 
    2. N: 前N行
返回数据: 返回一个表格的前N行
示例程序:
    >>> 排序后指标库 = sort(资产指标库, ['持仓市值'], '降序')
    >>> 前十大持仓资产 = limit(排序后指标库, 10)
'''


# 词法分析器：从语言文本解析程序Token串
from enum import Enum

# Token的类型
class TokenKind(Enum):
    Keyword = 1
    Identifier = 2
    StringLiteral = 3
    IntegerLiteral = 4
    DecimalLiteral = 5
    Seperator = 6
    Operator = 7
    EOF = 8
    Error = 9  # 新增错误类型


# Token的code
class Seperator(Enum):
    OpenBracket = 0  # [
    CloseBracket = 1  # ]
    OpenParen = 2  # (
    CloseParen = 3  # )
    OpenBrace = 4  # {
    CloseBrace = 5  # }
    Colon = 6  # :
    SemiColon = 7  # ;


# 运算符
class Op(Enum):
    Plus = 0  # +
    Minus = 1  # -
    Multiply = 3  # *
    Divide = 4  # /
    Comma = 5  # ,
    Dot = 6  # .
    Assign = 7  # =
    In = 8


class KeyWord(Enum):
    Function = 0


# 记录Token在源代码中的位置，方便报错和调试
class Position:
    def __init__(self, begin, end, line, col):
        self.begin = begin
        self.end = end
        self.line = line
        self.col = col

    def __str__(self):
        return '(行号: {}, 列号: {}, 位置: {})'.format(self.line, self.col, self.begin)


# 代表一个Token的数据结构
class Token:
    def __init__(self, kind, text, pos, code=None):
        self.kind = kind
        self.text = text
        self.pos = pos
        self.code = code

    def __str__(self):
        return 'Token @ {} \t {} \t {} \t{}'.format(self.pos, self.kind, self.text, self.code)


'''
 * 一个字符串流。其操作为：
 * peek():预读下一个字符，但不移动指针；
 * next():读取下一个字符，并且移动指针；
 * eof():判断是否已经到了结尾。
'''


class CharStream:
    def __init__(self, data):
        # 将输入程序的开始和末尾去掉及增加回车
        if data[-1] != '\n':  # 如果字符串末尾不能于回车，加上一个
            data = data + '\n'
        d = data[0]
        i = 0
        while d == '\n':
            i = i + 1
            d = data[i]
        data = data[i:]
        self.data = data
        self.pos = 0
        self.line = 1
        self.col = 1

    def peek(self):
        if len(self.data) <= self.pos:
            return None
        else:
            return self.data[self.pos]

    def next(self):
        if self.pos >= len(self.data):  # 检查是否越界
            return ''
        ch = self.data[self.pos]
        if self.pos > 0 and self.pos - 1 < len(self.data):
            ch_pre = self.data[self.pos - 1]
        else:
            ch_pre = ch
        if ch_pre == '\n':
            self.line = self.line + 1
            self.col = 1
        else:
            self.col = self.col + 1
        self.pos = self.pos + 1
        return ch

    def eof(self):
        return self.peek() is None

    # def get_position(self):
    #     return Position(self.pos + 1, self.pos + 1, self.line, self.col)

    def get_position(self):
        return Position(self.pos, self.pos, self.line, self.col)


'''
 * 词法分析器。
 * 词法分析器的接口像是一个流，词法解析是按需进行的。
 * 支持下面两个操作：
 * next(): 返回当前的Token，并移向下一个Token。
 * peek(): 预读当前的Token，但不移动当前位置。
 * peek2(): 预读第二个Token。
'''


class Scanner:
    KeyWords = {'function': KeyWord.Function}

    def __init__(self, stream):
        self.stream = stream
        self.tokens = list()
        self.lastPos = Position(0, 0, 0, 0)

    def next(self):
        if len(self.tokens) > 0:
            t = self.tokens.pop(0)
            if t is None:
                return self.get_token()
            else:
                return t
        else:
            return self.get_token()

    def peek(self):
        if len(self.tokens) > 0:
            t = self.tokens[0]
            if t is None:
                t = self.get_token()
                self.tokens.append(t)
        else:
            t = self.get_token()
            self.tokens.append(t)
        return t

    def peek2(self):
        if len(self.tokens) > 1:
            t = self.tokens[1]
            if t is None:
                while t is None:
                    self.tokens.append(self.get_token())
                    t = self.tokens[1]
        else:
            t = None
            while t is None:
                self.tokens.append(self.get_token())
                t = self.tokens[1]
        return t

    def get_next_pos(self):
        return self.peek().pos

    def get_last_pos(self):
        return self.lastPos

    @staticmethod
    def is_white_space(ch):
        return ch == ' ' or ch == '\t'

    @staticmethod
    def is_letter_digit_or_underscore(ch):
        return str.isdigit(ch) or str.isalpha(ch) or ch == '_' or ch == '（' or ch == '）'

    def skip_white_spaces(self):
        while self.is_white_space(self.stream.peek()):
            self.stream.next()

    # 字符串自变量
    def parse_string_literal(self, quotation_mark):
        pos = self.stream.get_position()
        token = Token(TokenKind.StringLiteral, '', pos)
        self.stream.next()  # 跳过开始的引号

        while True:
            ch = self.stream.peek()
            if ch is None or ch == '':  # 检查 EOF 或空字符
                token.kind = TokenKind.Error
                token.text = f'以{quotation_mark}开始的语句没有使用{quotation_mark}进行结尾'
                break
            elif ch == quotation_mark:
                self.stream.next()  # 跳过结束引号
                break
            else:
                next_char = self.stream.next()
                if next_char:  # 确保不是空字符串
                    token.text += next_char
                else:
                    token.kind = TokenKind.Error
                    token.text = f'Unexpected end of string literal'
                    break

        token.pos.end = self.stream.pos
        return token

    # 解析标识符
    def parse_identifier(self):
        pos = self.stream.get_position()
        token = Token(TokenKind.Identifier, '', pos)
        token.text += self.stream.next()
        while not self.stream.eof() and self.is_letter_digit_or_underscore(self.stream.peek()):
            token.text += self.stream.next()
        # pos.end = self.stream.pos + 1
        token.pos.end = self.stream.pos
        # 关键字是一种特殊的标识符
        if token.text in Scanner.KeyWords:
            token.kind = TokenKind.Keyword
            token.code = Scanner.KeyWords[token.text]
        if token.text == 'in':
            token.kind = TokenKind.Operator
            token.code = Op.In
        return token

    def get_token(self):
        pos = self.stream.get_position()
        self.skip_white_spaces()
        if self.stream.eof():
            token = Token(TokenKind.EOF, 'EOF', pos)
        else:
            ch = self.stream.peek()
            if str.isalpha(ch) or ch == '_':
                token =  self.parse_identifier()
            elif ch == '"' or ch == "'":
                token =  self.parse_string_literal(ch)
            elif ch == '(':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.OpenParen)
            elif ch == ')':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.CloseParen)
            elif ch == '[':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.OpenBracket)
            elif ch == ']':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.CloseBracket)
            elif ch == '{':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.OpenBrace)
            elif ch == '}':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.CloseBrace)
            elif ch == ':':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.Colon)
            elif ch == ',':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Op.Comma)
            elif ch == ';':
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.SemiColon)
            elif ch == '\n':  # 如果是一个空行
                self.stream.next()
                token = Token(TokenKind.Seperator, ch, pos, Seperator.SemiColon)

            # 解析数字
            elif str.isdigit(ch):
                self.stream.next()
                ch_num = int(ch)
                ch1 = self.stream.peek()
                literal = ''
                if ch_num == 0:
                    if str.isdigit(ch1):
                        print('0 cannot be followed by other digit now, at line: ' + str(
                            self.stream.line) + ' col: ' + str(self.stream.col))
                        # 先暂时跳过去
                        self.stream.next()
                        token = self.get_token()
                    else:
                        literal = '0'
                elif 1 <= ch_num <= 9:
                    literal += ch
                    while str.isdigit(ch1):
                        ch = self.stream.next()
                        literal += ch
                        ch1 = self.stream.peek()
                if ch1 == '.':
                    literal += '.'
                    self.stream.next()
                    ch1 = self.stream.peek()
                    while str.isdigit(ch1):
                        ch = self.stream.next()
                        literal += ch
                        ch1 = self.stream.peek()
                    # 返回一个浮点型token
                    # pos.end = self.stream.pos + 1
                    token = Token(TokenKind.DecimalLiteral, literal, pos)
                    token.pos.end = self.stream.pos
                    # return Token(TokenKind.DecimalLiteral, literal, pos)
                else:
                    # 返回一个整型token
                    token = Token(TokenKind.IntegerLiteral, literal, pos)
            elif ch == '.':
                self.stream.next()
                ch1 = self.stream.peek()
                if str.isdigit(ch1):
                    literal = '.'
                    while str.isdigit(ch1):
                        ch = self.stream.next()
                        literal += ch
                        ch1 = self.stream.peek()
                    # pos.end = self.stream.pos + 1
                    token = Token(TokenKind.DecimalLiteral, literal, pos)
                    token.pos.end = self.stream.pos
                    # pos.end = self.stream.pos + 1
                    # return Token(TokenKind.DecimalLiteral, literal, pos)
                else:
                    token = Token(TokenKind.Operator, '.', pos, Op.Dot)

            elif ch == '/':
                self.stream.next()
                token = Token(TokenKind.Operator, '/', pos, Op.Divide)
            elif ch == '+':
                self.stream.next()
                token = Token(TokenKind.Operator, '+', pos, Op.Plus)
            elif ch == '-':
                self.stream.next()
                token = Token(TokenKind.Operator, '-', pos, Op.Minus)
            elif ch == '*':
                self.stream.next()
                token = Token(TokenKind.Operator, '*', pos, Op.Multiply)
            elif ch == '=':
                self.stream.next()
                token = Token(TokenKind.Operator, '=', pos, Op.Assign)
            else:
                # 暂时去掉不能识别的字符
                print("Unrecognized pattern meeting:'" + ch + "', at ln:"
                      + str(self.stream.line) + " col: " + str(self.stream.col))
                self.stream.next()
                token = self.get_token()

        self.lastPos = token.pos
        return token



demos = list()

demo1 = '''
产品基本信息 = 指标('产品基本信息（时点）', ['产品代码', '产品简称'], {'PORT_TAGS': ['现金管理类产品'], 'START_DATE': '2023-03-01', 'END_DATE': '2023-03-01'})
'''
demo2 = '''
产品估值数据 = 指标('产品估值（时序）', ['产品代码', '统计日期', '总资产', '净资产'], {'PORT_TAGS': '现金管理类产品', 'START_DATE': '2023-03-01', 'END_DATE': '2024-03-01'})
'''
demo3 = '''
产品估值数据 = 指标('产品投资人（时序）', ['产品代码', '统计日期', '总份额', '前10大客户份额占比'], {'PORT_TAGS': ['现金管理类'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-06-17'})
'''
demo4 = '''
产品估值数据 = 指标('大类资产占比（时序）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-06-17'})
'''
demo5 = '''
产品估值数据 = 指标('申赎信息（时序）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-12', 'LOOP_DAYS': 30, 'METHOD': 'MAX'})
'''
demo6 = '''
产品估值数据 = 指标('资产基本信息（时点）', [], {'SEC_TAGS': ['信用债'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02'})
'''
demo7 = '''
产品估值数据 = 指标('持仓指标（新）', ['产品代码', '资产代码', '有效日期', '全价市值'], {'PORT_TAGS': ['现金管理类产品'], 'START_DATE': '2023-03-10', 'END_DATE': '2023-03-10', 'PENETRATE_TYPE': '01'})
'''
demo8 = '''
资产期间收益 = 指标('资产收益率（时序）', [], {'PORT_TAGS': ['现金管理类产品'], 'SEC_TAGS': ['资产管理产品', '企业债券', '信用债'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-04'})
'''
demo9 = '''
资产期间收益 = 指标('累计单利收益率', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02'})
'''
demo10 = '''
资产期间收益 = 指标('复权复利收益率', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02'})
'''
demo11 = '''
资产期间收益 = 指标('基准收益率', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02'})
'''
demo12 = '''
资产期间收益 = 指标('产品期间收益（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2023-03-01', 'END_DATE': '2023-03-15', 'BENCHMARK_ID': '', 'BENCHMARK_TYPE': '03'})
'''
demo13 = '''
资产期间收益 = 指标('产品费前收益率（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-04', 'BENCHMARK_ID': '000300.SH', 'BENCHMARK_TYPE': '03'})
'''
demo14 = '''
资产期间收益 = 指标('产品期间收益时点指标库', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02'})
'''
demo15 = '''
资产期间收益 = 指标('风险调整后收益（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-31', 'BENCHMARK_TYPE': '03'})
'''
demo16 = '''
资产期间收益 = 指标('产品每日收益（时序）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-31', 'BENCHMARK_TYPE': '03'})
'''
demo17 = '''
资产期间收益 = 指标('收益分解(产品)（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-04', 'PENETRATE_TYPE': '01'})
'''
# 下面这个有问题
demo18 = '''
资产期间收益 = 指标('收益分解(全维度)', [], {'PORT_TAGS': ['公募产品', '混合类产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-06-30', 'V_DATE': '2022-01-01', 'AGG': 'False', 'METHOD': 'CF'})
'''
demo19 = '''
资产期间收益 = 指标('产品收益（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-04', 'BENCHMARK_ID': '000300.SH', 'BENCHMARK_TYPE': '03'})
'''
demo20 = '''
资产期间收益 = 指标('产品内部基准收益', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-04'})
'''
demo21 = '''
资产期间收益 = 指标('产品期间收益率（时段）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-05-02', 'BENCHMARK_TYPE': '03'})
'''
demo22 = '''
资产期间收益 = 指标('产品Campisi归因', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2023-03-01', 'END_DATE': '2023-03-31', 'ABS': 'False'})
'''
demo23 = '''
资产期间收益 = 指标('市场风险（时序）', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-06-04'})
'''
demo24 = '''
资产期间收益 = 指标('产品VaR', [], {'PORT_TAGS': ['公募产品', '混合类产品'], 'START_DATE': '2022-05-02', 'END_DATE': '2022-06-04'})
'''
demo25 = '''
资产期间收益 = 指标('产品ECL', [], {'PORT_TAGS': ['开放式公募产品', '混合类产品'], 'DATA_DATE': '2023-03-31'})
'''
demo26 = '''
资产期间收益 = 指标('利率风险(产品)（时点）', [], {'PORT_TAGS': ['公募产品', '混合类产品'], 'START_DATE': '2023-03-01', 'END_DATE': '2023-03-05'})
'''
demo27 = '''
资产期间收益 = 指标('损益分析', [], {'PORT_TAGS': ['开放式公募产品'], 'START_DATE': '2023-03-01', 'END_DATE': '2023-03-31'})
'''
# 将所有demo添加到demos数组中
demos.append(demo1)
demos.append(demo2)
demos.append(demo3)
demos.append(demo4)
demos.append(demo5)
demos.append(demo6)
demos.append(demo7)
demos.append(demo8)
demos.append(demo9)
demos.append(demo10)
demos.append(demo11)
demos.append(demo12)
demos.append(demo13)
demos.append(demo14)
demos.append(demo15)
demos.append(demo16)
demos.append(demo17)
demos.append(demo18)
demos.append(demo19)
demos.append(demo20)
demos.append(demo21)
demos.append(demo22)
demos.append(demo23)
demos.append(demo24)
demos.append(demo25)
demos.append(demo26)
demos.append(demo27)
def get_demos():
    return demos


from muse.scanner import Scanner, CharStream, TokenKind
from muse.ast_struct import AstDumper, Prog
from muse.interpreter import Interpreter
from muse.resolver import Resolver
from muse.parser import Parser
import muse.engines.rule_engine as rule_engine
import muse.sample as sample
import muse.middata_sample as m_sample
from muse.auto_complete import parse_info
from muse.data_source.base import DataSourceInterface
from muse.data_source.ds_manager import ds_manager
import traceback
from collections import OrderedDict
import polars as pl
import re
SEP_LINE = '\n' + '-' * 100


# 词法分析
def token_analyse(program: str):
    # print('源代码: ' + SEP_LINE)
    # print(program)
    # print('词法分析结果: ' + SEP_LINE)
    scanner = Scanner(CharStream(program))
    tokens = list()
    while scanner.peek().kind != TokenKind.EOF:
        tokens.append(scanner.next())
    # for t in tokens:
    #     print(t)
    scanner = Scanner(CharStream(program))
    return scanner


# 语法分析
def ast_build(scanner: Scanner):
    parser = Parser(scanner)
    prog = parser.parse_prog()
    # print('\n语法分析后的AST: ' + SEP_LINE)
    ast_dumper = AstDumper()
    ast_dumper.visit(prog, '')
    return prog, parser.errors, ast_dumper.statements


# 语法消融及编译器错误排查
def resolve(prog: Prog):
    # print('\n编译器前检查: ' + SEP_LINE)
    resolver = Resolver()
    resolver.visit(prog, '')
    return resolver.props, resolver.errors, resolver.prompts


def run_program(prog: Prog, props: list):
    # print('\n执行AST: ' + SEP_LINE)
    interpreter = Interpreter(props)
    interpreter.visit(prog, '')
    # 增加监控规则的处理逻辑
    results = interpreter.results
    return results, interpreter.errors


def compile_and_run(program: str, ast_only=False, run_only=True):
    try:
        scanner = token_analyse(program)
        prog, parse_errors, statements = ast_build(scanner)
    except Exception:
        traceback.print_exc()
        return {'code': 'FAIL', 'data': ['语法解析出错，问题可能原因: 1) 包含中文的引号或其他特殊符号；2) 程序以非法结束符；']}

    # 如果只是想获得AST
    if ast_only:
        if len(parse_errors) > 0:
            p_errors = [{'error_pos': str(error.begin_pos), 'error_msg': error.msg} for error in parse_errors]
            return {'code': 'FAIL', 'parse_errors': p_errors, 'statements': statements}
        else:
            props, compile_errors, repositories = resolve(prog)
            if len(compile_errors) > 0:
                c_errors = [{'error_pos': str(error.begin_pos), 'error_msg': error.msg} for error in compile_errors]
                return {'code': 'FAIL', 'parse_errors': c_errors, 'statements': statements}
            else:
                return {'code': 'SUCCESS', 'repositories': repositories, 'statements': statements}
    # 如果是完全编译和运行程序
    else:
        # 如果AST分析时没有报错
        if len(parse_errors) == 0:
            props, compile_errors, repositories = resolve(prog)
            # 如果编译时没有报错
            if len(compile_errors) == 0:
                results, run_errors = run_program(prog, props)
                # 如果运行时没有问题
                # if len(run_errors) == 0:
                    # 变换之后的信息
                result_dict = rule_engine.parse_result(results)
                if run_only:
                    val = list(result_dict.values())[-1]
                    result_dict = dict()
                    result_dict['结果'] = val
                else:
                    result_dict = OrderedDict(reversed(result_dict.items()))
                error_msgs = [str(error) for error in run_errors]
                return {'code': 'SUCCESS', 'data': result_dict, 'error_msgs': error_msgs}
                # else:
                #     return {'code': 'FAIL', 'data': run_errors}

            else:
                error_msgs = [str(error) for error in compile_errors]
                return {'code': 'FAIL', 'data': error_msgs, 'error_msgs': error_msgs}
        else:
            error_msgs = [str(error) for error in parse_errors]
            return {'code': 'FAIL', 'data': error_msgs, 'error_msgs': error_msgs}


def run(program: str):
    ast_only = False
    run_only = True
    return compile_and_run(program, ast_only, run_only)


def trial(program: str):
    ast_only = False
    run_only = False
    return compile_and_run(program, ast_only, run_only)

def run_with_params(program: str, params: dict, head=0):
    # 先替换program中的所有参数
    appendix= ''
    for p_name, p_value in params.items():
        pattern = r"('{{p_name}}':\s*)(?:\[[^\]]*\]|'[^']*'|\{[^}]*\})(?=\s*[,}])"
        pattern = pattern.replace('{{p_name}}', p_name)
        replacement = rf"\1{p_name}"
        program = re.sub(pattern, replacement, program)
        if p_value == '' or p_value is None:
            appendix += f'{p_name} = ""\n'
        elif isinstance(p_value, str):
            appendix += f'{p_name} = "{p_value}"\n'
        else:
            appendix += f'{p_name} = {p_value}\n'

    full_program = appendix + program
    print('变换后的muse语句:')
    print(full_program)
    trial_result = trial(full_program)
    df_dict = dict()
    if trial_result['code'] == 'SUCCESS':
        result_data = trial_result['data']
        for k, df in result_data.items():
            if isinstance(df, dict):
                for kk, val in df.items():
                    if isinstance(val, pl.DataFrame):
                        if head == 0:
                            df_dict[kk] = val
                        else:
                            df_dict[kk] = val.head(head)
            elif isinstance(df, pl.DataFrame):
                if head == 0:
                    df_dict[k] = df
                else:
                    df_dict[k] = df.head(head)
    print(trial_result['error_msgs'])
    return df_dict


def compile_run_all():
    print(compile_and_run(sample.demo0))
    print(compile_and_run(sample.demo1))
    print(compile_and_run(sample.demo2))
    print(compile_and_run(sample.demo3))
    print(compile_and_run(sample.demo4))
    print(compile_and_run(sample.demo5))
    print(compile_and_run(sample.demo6))
    print(compile_and_run(sample.demo7))
    print(compile_and_run(sample.demo18))
    print(trial(sample.demo0))

def ast_check(program: str):
    ast_only = True
    run_only = True
    response = compile_and_run(program, ast_only=ast_only, run_only=run_only)

    statements = list()

    stms = response.get('statements', list())
    for s in stms:
        if '空语句' not in s:
            statements.append(s)

    colors, prompts = parse_info(statements)
    reps = response.get('repositories', list())
    for r in reps:
        colors['指标'] = colors['指标'] + r[1]

    response['prompts'] = prompts
    response['colors'] = colors
    if 'statements' in response:
        del response['statements']
    return response

def init_ds(data_source: DataSourceInterface):
    ds_manager.init_ds(data_source)

def get_ds()-> DataSourceInterface:
    return ds_manager.get_ds()

if __name__ == '__main__':
    print(trial(sample.demo1))
    print(trial(m_sample.demo2))
    ds = get_ds()
    print(ds.get_datasource_name())
    print(ds.get_all_repository_metadata())


# 消解器：将AST节点中的变量进行消解匹配和语法检查
from muse.ast_struct import AstVisitor, DotExp, Variable, FunctionCall, Binary
from muse.scanner import Op
import muse.data_interface as t


class Resolver(AstVisitor):
    def __init__(self):
        super().__init__()
        self.props = list()
        self.errors = list()
        self.results = dict()
        self.prompts = list()

    def visit_dot_exp(self, dot_exp: DotExp, additional):
        base_exp = self.visit(dot_exp.base_exp, additional)
        prop_exp = self.visit(dot_exp.prop, additional)

        prop_name = dot_exp.prop.name
        # 检查点运算后面访问的prop(指标)是不是合理的，能够获取的指标
        if t.is_valid_indicator(prop_name):
            self.props.append(prop_name)
        else:
            # error_msg = '点运算后面引用的指标 "{}" 当前不支持. 当前支持的指标有: {}'.format(prop_name, t.VALID_INDS)
            error_msg = '点运算后面引用的指标 "{}" 当前不支持.'.format(prop_name)
            # self.add_error(error_msg, dot_exp.end_pos)
        return prop_name

    def visit_function_call(self, func_call: FunctionCall, additional):
        # 获得函数调用名称
        name = func_call.name
        ps = list()
        for p in func_call.params:
            ps.append(self.visit(p, additional))

        # 检查函数是不是合理的
        if not t.is_valid_func(name):
            error_msg = '函数: "{}" 当前不支持.'.format(name)
            self.add_error(error_msg, func_call.begin_pos)
        elif name == 'agg':
            if len(ps) < 3:
                error_msg = '函数: agg 需要3个参数: 分别为金融对象, [分组字段], 分组方法'
                self.add_error(error_msg, func_call.begin_pos)
        elif name == '指标':
            if len(ps) < 3:
                error_msg = '函数: 指标 需要3个参数，分别为指标库名，指标列表，参数字典'
                self.add_error(error_msg, func_call.begin_pos)
            else:
                self.prompts.append(ps)
        # elif name == '资产' or name == '主体' or name == '资产':
        #     if len(ps) < 1:
        #         error_msg = '函数: "{}" 需要至少1个参数: [标签列表]'
        #         self.add_error(error_msg, func_call.begin_pos)

    def visit_binary(self, binary: Binary, additional):
        exp1 = binary.exp1
        exp2 = binary.exp2

        op = binary.op
        error_msg = 'in 操作符号前后两部分均应该为标签函数调用，"{}" 不是支持的标签调用。目前支持的标签为 "{}"'
        # 如果是in操作，要保证左侧和右侧都是函数调用
        if op == Op.In and isinstance(exp1, FunctionCall) and isinstance(exp2, FunctionCall):
            func1_name = exp1.name
            func2_name = exp2.name
            if func1_name not in t.TAGS:
                self.add_error(error_msg.format(func1_name, t.TAGS), exp1.begin_pos)
            if func2_name not in t.TAGS:
                self.add_error(error_msg.format(func2_name, t.TAGS), exp2.begin_pos)

        elif op == Op.In and isinstance(exp1, Binary) and isinstance(exp2, FunctionCall):
            b1 = exp1
            if exp1.op == Op.In and isinstance(b1.exp1, FunctionCall) and isinstance(b1.exp2, FunctionCall):
                func1_name = b1.exp1.name
                func2_name = b1.exp2.name
                if func1_name not in t.TAGS:
                    self.add_error(error_msg.format(func1_name, t.TAGS), b1.exp1.begin_pos)
                if func2_name not in t.TAGS:
                    self.add_error(error_msg.format(func2_name, t.TAGS), b1.exp2.begin_pos)
        else:
            v1 = self.visit(exp1, 'left')
            v2 = self.visit(exp2, None)
            if op == Op.Assign:
                self.results[v1] = v2
                # print('调用赋值表达式, 把 {} 放入 {}'.format(v1, self.results))

    def visit_variable(self, variable: Variable, additional):
        var_name = variable.name
        if additional is not None and additional == 'left':
            return var_name
        # else:
        #     if var_name not in self.results and not t.is_valid_indicator(var_name):
        #         error_msg = '变量 "{}" 没有被定义就使用.'.format(var_name)
        #         self.add_error(error_msg, variable.begin_pos)
        return var_name



import pathlib
import os
# MUSE所至目录
APP_LOCATE = str(pathlib.Path(os.path.abspath(__file__)).parent)
# 本地文件地址
DATA_PATH = APP_LOCATE + '/data'
# 数据中台地址
MID_BASE_URL = "http://jixunet.top/jxdm_zyrisk"
# 数据中台资产标签组名称
ASSET_TAG_GROUP = "压测标签组"
# 数据中台产品标签组名称
PORT_TAG_GROUP = "限额指标产品标签组"
# 数据中台指标库父级名称，可使用逗号表示多个开放的指标库
REPOSITORY_GROUP = "API指标库"


import polars as pl
from muse.data_source.ds_manager import ds_manager
from muse.func_desc import VALID_FUNCS

TAGS = ['资产', '主体', '产品']
# BASIC_COLS = dsg.get_basic_inds()
def get_basic_cols():
    dsg = ds_manager.get_ds()
    return dsg.get_basic_inds()

# ASSET_INDS = dsg.get_asset_tags()       # 资产标签
def get_asset_tags():
    dsg = ds_manager.get_ds()
    return dsg.get_asset_tags()

# HLD_INDS = dsg.get_asset_inds()          # 持仓指标
def get_asset_cols():
    dsg = ds_manager.get_ds()
    return dsg.get_asset_inds()

# ISSUER_INDS = ASSET_INDS + HLD_INDS
def get_asset_all_cols():
    return get_asset_tags() + get_asset_cols()

# PORT_TAGS = dsg.get_port_tags()       # 产品标签
def get_port_tags():
    dsg = ds_manager.get_ds()
    return dsg.get_port_tags()

# PORT_ZB = dsg.get_port_inds()               # 产品指标
def get_port_cols():
    dsg = ds_manager.get_ds()
    return dsg.get_port_inds()

# PORT_INDS = PORT_TAGS + PORT_ZB     # 产品标签+指标
def get_port_all_cols():
    return get_port_tags() + get_port_cols()

# VALID_INDS = ASSET_INDS + HLD_INDS + ISSUER_INDS + PORT_INDS
def get_all_valid_cols():
    return get_asset_all_cols() + get_port_all_cols()


def get_port_inds(run_method: str, start_date: str, end_date: str, port_ids, inds: list):
    dsg = ds_manager.get_ds()
    port_inds = dsg.get_data(run_method, "产品指标", port_ids = port_ids, start_date=start_date, end_date=end_date, inds=inds)

    cols = ['产品代码', '日期']
    if inds is not None and len(inds) > 0:
        prop_list = filter_inds('产品', inds)
        cols = cols + prop_list
    if len(port_ids) > 0:
        df_lazy = (
            port_inds.lazy().filter(pl.col('产品代码').is_in(port_ids)).select(tuple(cols))
        )
    else:
        df_lazy = (
            port_inds.lazy().select(tuple(cols))
        )
    return df_lazy.collect()

def get_ports(run_method: str, start_date: str, end_date: str, tags: list, port_ids: list, inds: list, port_inds=None):
    # 获取产品指标（假设返回的是 Polars DataFrame）
    dsg = ds_manager.get_ds()
    if port_inds is None:
        port_inds = dsg.get_data(run_method, "产品指标", port_ids = port_ids, start_date=start_date, end_date=end_date, tags=tags, inds=inds)
    if port_inds is None or port_inds.is_empty():
        return pl.DataFrame()
    # 1. 产品标签筛选（OR 关系：任意一个标签为 "是"）
    port_condition = None
    if tags:
        # 检查标签是否存在
        valid_port_tags = [tag for tag in tags if tag in port_inds.columns]
        if not valid_port_tags:
            raise ValueError(f"产品标签 {tags} 不在数据列中！")

        # 动态构建 OR 条件
        port_condition = pl.any_horizontal(
            *[pl.col(tag) == "是" for tag in valid_port_tags]
        )
    if port_condition is not None:
        filtered_df = port_inds.filter(port_condition)
    else:
        filtered_df = port_inds  # 无筛选条件

    return filtered_df


def get_asset_by_tags(run_method: str, start_date: str, end_date: str, tags: list, drop_duplicates=False):
    dsg = ds_manager.get_ds()
    hlds = dsg.get_data(run_method, "持仓指标", start_date=start_date, end_date=end_date, tags=tags)
    if hlds is None or hlds.is_empty():
        return pl.DataFrame()

    # 1. 资产标签筛选（OR 关系：任意一个标签为 "是"）
    ast_condition = None
    if tags:
        # 检查标签是否存在
        valid_ast_tags = [tag for tag in tags if tag in hlds.columns]
        if not valid_ast_tags:
            raise ValueError(f"资产标签 {tags} 不在数据列中！")

        # 动态构建 OR 条件
        ast_condition = pl.any_horizontal(
            *[pl.col(tag) == "是" for tag in valid_ast_tags]
        )

    if ast_condition is not None:
        filtered_df = hlds.filter(ast_condition)
    else:
        filtered_df = hlds  # 无筛选条件
    return filtered_df

def get_asset_by_ids(run_method: str, start_date: str, end_date: str, sec_ids: list):
    dsg = ds_manager.get_ds()
    ast = dsg.get_data(run_method, '持仓指标', sec_ids = sec_ids, start_date=start_date, end_date=end_date)

    if len(sec_ids) > 0:
        df_lazy = (
            ast.lazy().filter(pl.col('资产代码').is_in(sec_ids))
        )
    else:
        df_lazy = (
            ast.lazy().select('资产代码', '资产简称', '主体代码').unique()
        )
    result = df_lazy.collect()
    return result


def get_asset_inds(run_method: str, start_date: str, end_date: str, sec_ids: list, inds: list):
    dsg = ds_manager.get_ds()
    ast_inds = dsg.get_data(run_method, '资产指标', sec_ids = sec_ids, start_date=start_date, end_date=end_date)
    cols = ['资产代码', '日期']
    if inds is not None and len(inds) > 0:
        prop_list = filter_inds('资产', inds)
        cols = cols + prop_list
    if len(sec_ids) > 0:
        df_lazy = (
            ast_inds.lazy().filter(pl.col('资产代码').is_in(sec_ids)).select(tuple(cols))
        )
    else:
        df_lazy = (
            ast_inds.lazy().select(tuple(cols))
        )

    return df_lazy.collect()


def get_assets(run_method: str, start_date: str, end_date: str, tags: list, sec_ids: list, inds: list):
    # 获取持仓数据（假设返回的是 Polars DataFrame
    dsg = ds_manager.get_ds()
    hlds = dsg.get_data(run_method, "持仓指标", sec_ids = sec_ids, start_date=start_date, end_date=end_date, tags=tags, inds=inds)
    if hlds is None or hlds.is_empty():
        return pl.DataFrame()

    # 1. 资产标签筛选（OR 关系：任意一个标签为 "是"）
    ast_condition = None
    if tags:
        # 检查标签是否存在
        valid_ast_tags = [tag for tag in tags if tag in hlds.columns]
        if not valid_ast_tags:
            raise ValueError(f"资产标签 {tags} 不在数据列中！")

        # 动态构建 OR 条件
        ast_condition = pl.any_horizontal(
            *[pl.col(tag) == "是" for tag in valid_ast_tags]
        )
    if ast_condition is not None:
        filtered_df = hlds.filter(ast_condition)
    else:
        filtered_df = hlds  # 无筛选条件

    return filtered_df

def get_issuer_by_tags(start_date: str, end_date: str, tags: list, drop_duplicates=False):
    dsg = ds_manager.get_ds()
    issuer = dsg.get_data('主体信息')
    df_lazy = (
        issuer.lazy().filter(pl.col('主体标签').is_in(tags))
    )

    return df_lazy.collect()

def get_issuer_by_ids(start_date: str, end_date: str, issuer_ids: list):
    dsg = ds_manager.get_ds()
    issuer = dsg.get_data('主体信息')
    if len(issuer_ids) > 0:
        df_lazy = (
            issuer.lazy().filter(pl.col('主体代码').is_in(issuer_ids))
        )
    else:
        df_lazy = (
            issuer.lazy().select('主体代码', '主体名称').unique()
        )
    return df_lazy.collect()

def get_issuer_inds(run_method: str, start_date: str, end_date: str, issuer_ids: list, inds: list):
    dsg = ds_manager.get_ds()
    issuer_inds = dsg.get_data(run_method, '主体指标', start_date=start_date, end_date=end_date, inds=inds)
    cols = ['主体代码', '日期']
    if inds is not None and len(inds) > 0:
        prop_list = filter_inds('主体', inds)
        cols = cols + prop_list

    if len(issuer_ids) > 0:
        df_lazy =(
            issuer_inds.lazy().filter(pl.col('主体代码').is_in(issuer_ids)).select(tuple(cols))
        )
    else:
        df_lazy = (
            issuer_inds.lazy().select(tuple(cols))
        )
    return df_lazy.collect()

def get_issuers(run_method: str, start_date: str, end_date: str, tags, issuer_ids: list, inds: list):
    if len(issuer_ids) > 0:
        issuers = get_issuer_by_ids(start_date, end_date, issuer_ids)
    elif len(tags) > 0:
        issuers = get_issuer_by_tags(start_date, end_date, tags)
    else:
        issuers = get_issuer_by_ids(start_date, end_date, [])
    # 根据资产中的代码和<日期>, 查询指标数据
    issuer_list = issuers.select('主体代码').unique()
    issuer_inds = get_issuer_inds(run_method, start_date, end_date, issuer_list, inds)
    issuers = issuers.join(issuer_inds, on=['主体代码'], how='left')
    return issuers

def get_hld_inds(run_method: str, start_date: str, end_date: str, port_ids: list, sec_ids: list, inds: list, penetrate=None):
    dsg = ds_manager.get_ds()
    cols = ['资产代码', '产品代码', '日期']
    if inds is not None and len(inds) > 0:
        hld_list = filter_inds('持仓', inds)
        cols = cols + hld_list

    # 根据资产中的代码和<日期>, 查询指标数据
    hld = dsg.get_data(run_method, '持仓指标', port_ids = port_ids, sec_ids = sec_ids, start_date=start_date, end_date=end_date, inds = inds, penetrate = penetrate)

    df_lazy = (
        hld.lazy().select(tuple(cols)).filter(pl.col('产品代码').is_in(port_ids) & pl.col('资产代码').is_in(sec_ids))
    )
    return df_lazy.collect()

def get_hlds3(run_method: str, start_date: str, end_date: str, asts: list, ports: list, inds: list):
    sec_ids = asts.select('资产代码').unique()
    port_ids = ports.select('产品代码').unique()
    # 获取持仓指标
    hlds = get_hld_inds(run_method, start_date, end_date, port_ids, sec_ids, inds)
    df_lazy =(
       hlds.lazy().join(asts.lazy(), on=['资产代码', '日期'], how='left',suffix="_asts")
                  .join(ports.lazy(), on=['产品代码', '日期'], how='left',suffix="_ports")
    )
    hlds = df_lazy.collect()
    df_lazy = (
        hlds.lazy().select([col for col in hlds.columns if not ('_asts' in col or '_ports' in col)])
    )
    hlds = df_lazy.collect()
    return hlds

def get_hlds2(run_method: str, start_date: str, end_date: str, asts, issuers, ports, inds: list, penetrate=None):
    issuer_ids = issuers.select('主体代码').unique()
    ast_lazy = (
        asts.lazy().filter(pl.col('主体代码').is_in(issuer_ids)).join(issuers.lazy(), on=['主体代码', '日期'], how='left')
    )
    ast_issuers = ast_lazy.collect()
    sec_ids = ast_issuers.select('资产代码').unique()
    port_ids = ports.select('产品代码').unique()
    # 获取持仓指标
    hlds = get_hld_inds(run_method, start_date, end_date, port_ids, sec_ids, inds, penetrate=penetrate)
    # 将资产数据补齐
    df_lazy = (
        hlds.lazy().join(ast_lazy, on=['资产代码', '日期'], how='left').join(ports.lazy(), on=['产品代码', '日期'], how='left')
    )
    hlds = df_lazy.collect()
    return hlds

def get_hlds(run_method: str, start_date: str, end_date: str, ast_tags: list, port_tags: list,  inds: list, sec_ids=None, port_ids=None, penetrate=None):
    dsg = ds_manager.get_ds()
    # try:
        # 获取持仓数据（假设返回的是 Polars DataFrame）
    merge_tags = ast_tags + port_tags
    hlds = dsg.get_data(run_method, "持仓指标", sec_ids = sec_ids, port_ids = port_ids, start_date=start_date, end_date=end_date, tags=merge_tags, inds=inds, penetrate=penetrate)
    if hlds is None or hlds.is_empty():
        return pl.DataFrame()

    # 1. 资产标签筛选（OR 关系：任意一个标签为 "是"）
    ast_condition = None
    if ast_tags:
        # 检查标签是否存在
        valid_ast_tags = [tag for tag in ast_tags if tag in hlds.columns]
        if not valid_ast_tags:
            raise ValueError(f"资产标签 {ast_tags} 不在数据列中！")

        # 动态构建 OR 条件
        ast_condition = pl.any_horizontal(
            *[pl.col(tag) == "是" for tag in valid_ast_tags]
        )

    # 2. 产品标签筛选（OR 关系：任意一个标签为 "是"）
    port_condition = None
    if port_tags:
        # 检查标签是否存在
        valid_port_tags = [tag for tag in port_tags if tag in hlds.columns]
        if not valid_port_tags:
            raise ValueError(f"产品标签 {port_tags} 不在数据列中！")

        # 动态构建 OR 条件
        port_condition = pl.any_horizontal(
            *[pl.col(tag) == "是" for tag in valid_port_tags]
        )

    # 3. 组合条件（AND 关系）
    if ast_condition is not None and port_condition is not None:
        filtered_df = hlds.filter(ast_condition & port_condition)
    elif ast_condition is not None:
        filtered_df = hlds.filter(ast_condition)
    elif port_condition is not None:
        filtered_df = hlds.filter(port_condition)
    else:
        filtered_df = hlds  # 无筛选条件
    return filtered_df

def get_repository_inds(run_method: str, repository:str, ind_list:list, params:dict):
    dsg = ds_manager.get_ds()
    return dsg.get_repository_inds(run_method, repository, ind_list, params)

def get_inds(inds_type: str) -> list:
    dsg = ds_manager.get_ds()
    inds_df = dsg.get_data('测试', inds_type)
    return list(set(inds_df.columns) - set(get_basic_cols()))

def filter_inds(ind_type: str, inds: list):
    if len(inds) <= 0:
        return list()
    if ind_type == '资产':
        inds_list = get_asset_tags()
    elif ind_type == '主体':
        inds_list = get_asset_all_cols()
    elif ind_type == '持仓':
        inds_list = get_asset_cols()
    else:
        inds_list = get_port_all_cols()

    return list(set(inds).intersection(set(inds_list)))

def find_prop(cols: list):
    inds = [c for c in cols if c not in get_basic_cols()]
    if len(inds) > 0:
        return inds[0]
    else:
        return None

def find_all_props(cols: list):
    return [c for c in cols if c not in get_basic_cols()]

def find_basic(cols: list) -> list:
    inds = [c for c in cols if c in get_basic_cols()]
    return inds

def is_valid_indicator(prop: str) -> bool:
    return prop in get_all_valid_cols() + get_basic_cols()

def is_valid_ind(prop: str) -> bool:
    return prop in get_all_valid_cols()

def is_valid_func(func_name: str) -> bool:
    return func_name in VALID_FUNCS


if __name__ == '__main__':
    port_list = ['P01', 'P02', 'P03']
    port_props = ['单位净值', '个人投资者占比']
    port_tags = ['公募开放式', '固定收益类']
    ports = get_ports('', '', port_tags, [], port_props)
    asts = get_assets('', '', ['信用债', '债券回购'], [], ['资产到期日', '久期'])
    # df = get_issuers('', '', ['银行金融机构', '非银行金融机构'], [], ['上一季度末净资产'])
    sec_list = ['B1', 'B7', 'B8']
    # df = get_hld_inds('', '', port_list, sec_list, ['持仓市值'])
    df = get_hlds('', '', asts, ports, ['持仓市值'])
    df = df.select(pl.col('单位净值').alias('复权单位净值'))
    print(df)



from collections import OrderedDict
import polars as pl
import traceback
import re
import logging

def extract_fields(expression):
    # 匹配中文字符、英文字母、数字和下划线组成的字段名
    # pattern = r'([\u4e00-\u9fa5a-zA-Z0-9_（）()+-]+)\s*(?:<|>|==|!=|<=|>=)\s*'
    pattern = r'(?:\(*([\u4e00-\u9fa5a-zA-Z0-9_（）+-]+)\)*)\s*(?:<|>|==|!=|<=|>=)\s*'
    fields = re.findall(pattern, expression)
    # 去重并返回
    return list(set(fields))

def use_rule(rules: dict, df: pl.DataFrame) -> pl.DataFrame:
    if df is None or df.is_empty():
        return df
    result_dfs = []
    monitor_rule = list(rules.values())[0]
    # 将每个rule中需要的字段都提取出来
    all_filter_fields = list()
    for _, rule in rules.items():
        fields = extract_fields(rule)
        all_filter_fields = all_filter_fields + fields

    all_filter_fields = list(set(all_filter_fields))
    normal_df = df
    # 将空值和inf过滤出来
    for col in all_filter_fields:
        try:
            inf_df = normal_df.filter(pl.col(col).is_infinite())
        except:
            inf_df = pl.DataFrame()
        if not inf_df.is_empty():
            inf_df_after_rule = inf_df.with_columns(pl.lit('除0异常').alias('监控结果'), pl.lit(monitor_rule).alias('监控规则'))
            normal_df = normal_df.filter(~(pl.col(col).is_infinite()))
            result_dfs.append(inf_df_after_rule)
        try:
            none_df = normal_df.filter(pl.col(col).is_null() | pl.col(col).is_nan())
        except:
            none_df = pl.DataFrame()
        if not none_df.is_empty():
            none_df_after_rule = none_df.with_columns(pl.lit('其他异常').alias('监控结果'), pl.lit(monitor_rule).alias('监控规则'))
            normal_df = normal_df.filter(~(pl.col(col).is_null() | pl.col(col).is_nan()))
            result_dfs.append(none_df_after_rule)
    try:
        expressions = {
            rule_result: build_expression(rule_expr, all_filter_fields) for rule_result, rule_expr in rules.items()
        }
    except Exception:
        logging.info(traceback.format_exc())
        return df.with_columns(
                pl.lit('监控规则出现错误：请检查是否在监测规则中引用了没有的字段。').alias("监控结果")
        )

    for rule_result, mask in expressions.items():
        try:
            # 直接使用生成的 mask
            filtered_df = normal_df.filter(mask[0])
            filtered_df = filtered_df.with_columns(
                pl.lit(rule_result).alias("监控结果"),
                # pl.lit(monitor_rule) # 所有同一规则
                pl.lit(mask[1]).alias("监控规则")  # 不同规则
            )
            result_dfs.append(filtered_df)
        except Exception:
            logging.error(traceback.format_exc())
            return df.with_columns(
                pl.lit('监控规则出现错误：请检查是否在监测规则中引用了没有的字段。').alias("监控结果")
            )

    if result_dfs:
        sorted_col = all_filter_fields[0]
        result = pl.concat(result_dfs)
        # 对result进行排序
        custom_order = ['触发超限', '触发预警', '除0异常', '其他异常', '未达到条件', '正常']
        existing_values = result["监控结果"].unique().to_list()

        # 合并custom_order和未定义值（去重）
        full_order = custom_order + [v for v in existing_values if v not in custom_order]
        result = result.with_columns(
            pl.col("监控结果")
            .cast(pl.Enum(full_order))
        ).sort("监控结果", sorted_col, descending=[False, True])
        return result
    else:
        return df


def build_expression(rule_str: str, df_columns: list) -> pl.Expr:
    # 构造表达式
    rule_logic = rule_str
    for col in sorted(df_columns, key=len, reverse=True):
        rule_logic = rule_logic.replace(col, f'pl.col("{col}")')

        # 使用 Polars 表达式解析
    return eval(rule_logic), rule_str

def get_result_col(columns):
    score_list = list()
    for col in columns:
        score = 0
        if '-' in str(col):
            score += 1
        if '+' in str(col):
            score += 1
        if '*' in str(col):
            score += 1
        if '/' in str(col):
            score += 1
        score_list.append(score)

    result_col = None
    if len(score_list):
        max_value = max(score_list)  # 最大值
        if max_value > 0:
            max_index = score_list.index(max_value)  # 第一个最大值的索引
            result_col = columns[max_index]
    return result_col


def parse_result(results: OrderedDict):
    r = OrderedDict()
    rules = dict()
    for k, v in results.items():
        if k == '监控规则':
            rules = v
        else:
            r[k] = v

    if len(r) > 0 and len(rules) > 0:
        name, df = r.popitem()
        r[name] = use_rule(rules, df)
    return r


# AST结构定义
from muse.scanner import Position, Op
from muse.error import CompileError
import logging


# AST节点基类
class AstNode:
    def __init__(self, begin_pos: Position, end_pos: Position, is_error=False):
        self.begin_pos = begin_pos
        self.end_pos = end_pos
        self.is_error = is_error

    # visitor模式中, 用于接受visitor的访问
    def accept(self, visitor, additional) :
        pass


# 声明: 所有声明都会对应一个符号
class Decl(AstNode):
    def __init__(self, begin_pos: Position, end_pos: Position, name: str, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.name = name


##### 程序语句开始
class Statement(AstNode):
    def __init__(self, begin_pos: Position, end_pos: Position, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)


# 调用签名
class ParameterList(AstNode):
    def __init__(self, begin_pos: Position, end_pos: Position, params: list, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.params = params

    def __str__(self):
        return '参数列表'

    def accept(self, visitor, additional) :
        return visitor.visit_parameter_list(self, additional)


# 程序块
class Block(Statement):
    def __init__(self, begin_pos: Position, end_pos: Position, statements: list, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.statements = statements

    def __str__(self):
        return '程序块'

    def accept(self, visitor, additional):
        return visitor.visit_block(self, additional)


# 整个程序
class Prog(Block):
    def __init__(self, begin_pos: Position, end_pos: Position, statements: list, is_error=False):
        super().__init__(begin_pos, end_pos, statements, is_error)
        self.statements = statements

    def __str__(self):
        return '程序整体'

    def accept(self, visitor, additional) :
        return visitor.visit_prog(self, additional)


##### 表达式开始
class Expression(AstNode):
    def __init__(self, begin_pos: Position, end_pos: Position, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)


# 变量引用
class Variable(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, name: str, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.name = name

    def __str__(self):
        return self.name

    def accept(self, visitor, additional) :
        return visitor.visit_variable(self, additional)


# 二元表达式
class Binary(Expression):
    def __init__(self, op: Op, exp1: Expression, exp2: Expression, is_error=False):
        super().__init__(exp1.begin_pos, exp2.end_pos, is_error)
        self.op = op
        self.exp1 = exp1
        self.exp2 = exp2
        self.exp1.parent_node = self
        self.exp2.parent_node = self

    def __str__(self):
        return '二元表达式: ' + self.op.name

    def accept(self, visitor, additional) :
        return visitor.visit_binary(self, additional)


# 一元表达式
class Unary(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, op: Op, exp: Expression, is_prefix, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.op = op
        self.exp = exp
        self.is_prefix = is_prefix
        self.exp.parent_node = self

    def __str__(self):
        return '一元表达式: ' + self.op.name

    def accept(self, visitor, additional) :
        return visitor.visit_unary(self, additional)


# 函数调用
class FunctionCall(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, name: str, params: list, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.name = name  # 函数名
        self.params = params  # 函数参数列表
        for v in self.params:
            v.parent_node = self

    def __str__(self):
        return '函数调用: ' + self.name

    def accept(self, visitor, additional) :
        return visitor.visit_function_call(self, additional)


# 字符串字面值
class StringLiteral(Expression):
    def __init__(self, pos: Position, value: str, is_error=False):
        super().__init__(pos, pos, is_error)
        self.value = value
        self.the_type = '字符串'

    def accept(self, visitor, additional) :
        return visitor.visit_string_literal(self, additional)


# 整型字面值
class IntegerLiteral(Expression):
    def __init__(self, pos: Position, value: int, is_error=False):
        super().__init__(pos, pos, is_error)
        self.value = value
        self.the_type = '整数'

    def accept(self, visitor, additional) :
        return visitor.visit_integer_literal(self, additional)


# 实数字面值
class DecimalLiteral(Expression):
    def __init__(self, pos: Position, value: float, is_error=False):
        super().__init__(pos, pos, is_error)
        self.value = value
        self.the_type = '浮点数'

    def accept(self, visitor, additional) :
        return visitor.visit_decimal_literal(self, additional)


# 点表达式
class DotExp(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, base_exp: Expression, prop: Expression, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.base_exp = base_exp
        self.prop = prop
        self.base_exp.parent_node = self
        self.prop.parent_node = self

    def __str__(self):
        return '点表达式'

    def accept(self, visitor, additional) :
        return visitor.visit_dot_exp(self, additional)


# 数组字面量
class ArrayLiteral(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, exps: list, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.exps = exps

    def __str__(self):
        return '数组'

    def accept(self, visitor, additional) :
        return visitor.visit_array_literal(self, additional)


class MapLiteral(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position, exps: list, is_error=False):
        super().__init__(begin_pos, end_pos, is_error)
        self.exps = exps

    def accept(self, visitor, additional) :
        return visitor.visit_map_literal(self, additional)


# 错误表达式
class ErrorExp(Expression):
    def __init__(self, begin_pos: Position, end_pos: Position):
        super().__init__(begin_pos, end_pos, True)
        self.is_error = True

    def accept(self, visitor, additional):
        return visitor.visit_error_expression(self, additional)


# 变量声明
class VariableDeclare(Decl):
    def __init__(self, begin_pos: Position, end_pos: Position, name: str, exp: Expression, is_error=False):
        super().__init__(begin_pos, end_pos, name, is_error)
        self.exp = exp

    def __str__(self):
        return '变量声明: ' + self.name

    def accept(self, visitor, additional) :
        return visitor.visit_variable_declare(self, additional)


# 变量声明语句
class VariableStatement(Statement):
    def __init__(self, begin_pos: Position, end_pos: Position, variable_declare: VariableDeclare, is_error=False):
        super().__init__(begin_pos, end_pos, variable_declare, is_error)
        self.variable_declare = variable_declare

    def __str__(self):
        return '变量声明语句'

    def accept(self, visitor, additional) :
        return visitor.visit_variable_statement(self, additional)


# 表达式语句
class ExpressionStatement(Statement):
    def __init__(self, end_pos: Position, exp: Expression, is_error=False):
        super().__init__(exp.begin_pos, end_pos, is_error)
        self.exp = exp
        self.parent_node = self

    def __str__(self):
        return '表达式语句'

    def accept(self, visitor, additional) :
        return visitor.visit_expression_statement(self, additional)


class ErrorStatement(Statement):
    def __init__(self, begin_pos: Position, end_pos: Position):
        super().__init__(begin_pos, end_pos, True)
        self.is_error = True

    def accept(self, visitor, additional) :
        return visitor.visit_error_statement(self, additional)


class EmptyStatement(Statement):
    def __init__(self, pos: Position, is_error=False):
        super().__init__(pos, pos, is_error)

    def __str__(self):
        return '空语句'

    def accept(self, visitor, additional) :
        return visitor.visit_empty_statement(self, additional)


#  * 对AST做遍历的Vistor。
#  * 这是一个基类，定义了缺省的遍历方式。子类可以覆盖某些方法，修改遍历方式。
class AstVisitor:
    def __init__(self):
        self.errors = list()
        self.warnings = list()

    def add_error(self, msg: str, pos: Position):
        self.errors.append(CompileError(msg, pos, False))
        logging.info('@{} : {}'.format(pos, msg))

    def add_warning(self, msg: str, pos: Position):
        self.warnings.append(CompileError(msg, pos, True))
        logging.info('@{} : {}'.format(pos, msg))

    def visit(self, ast_node: AstNode, additional=None):
        return ast_node.accept(self, additional)

    def visit_prog(self, prog: Prog, additional=None):
        return self.visit_block(prog, additional)

    def visit_block(self, block: Block, additional=None):
        result = None
        for s in block.statements:
            result = self.visit(s, additional)
        return result

    def visit_parameter_list(self, param_list: ParameterList, additional=None):
        ps = list()
        for p in param_list.params:
            ps.append(self.visit(p, additional))
        return ps

    def visit_variable_declare(self, var_declare: VariableDeclare, additional):
        if var_declare.exp is not None:
            return self.visit(var_declare.exp, additional)

    def visit_variable_statement(self, var_statement: VariableStatement, additional):
        return self.visit(var_statement.variable_declare, additional)

    def visit_variable(self, variable: Variable, additional):
        return None

    def visit_expression_statement(self, statement: ExpressionStatement, additional):
        return self.visit(statement.exp, additional)

    def visit_error_statement(self, statement: ErrorStatement, additional):
        return None

    def visit_empty_statement(self, statement: EmptyStatement, additional):
        return None

    def visit_binary(self, binary: Binary, additional):
        self.visit(binary.exp1, additional)
        self.visit(binary.exp2, additional)

    def visit_unary(self, unary: Unary, additional):
        self.visit(unary.exp, additional)

    def visit_function_call(self, func_call: FunctionCall, additional):
        for param in func_call.params:
            self.visit(param, additional)
        return None

    def visit_string_literal(self, literal: StringLiteral, additional):
        return literal.value

    def visit_integer_literal(self, literal: IntegerLiteral, additional):
        return literal.value

    def visit_decimal_literal(self, literal: DecimalLiteral, additional):
        return literal.value

    def visit_dot_exp(self, dot_exp: DotExp, additional):
        self.visit(dot_exp.base_exp, additional)
        self.visit(dot_exp.prop, additional)

    def visit_array_literal(self, array: ArrayLiteral, additional):
        values = list()
        for exp in array.exps:
            values.append(self.visit(exp, additional))
        return values

    def visit_map_literal(self, map_exp: MapLiteral, additional):
        values = dict()
        for exp in map_exp.exps:
            k = self.visit(exp['key'], additional)
            v = self.visit(exp['val'], additional)
            values[k] = v
        return values

    def visit_error_expression(self, error_exp: ErrorExp, additional):
        return None


blanks = '@'
# blanks = ''

# 打印AST的调试程序
class AstDumper(AstVisitor):
    def __init__(self):
        self.statements = list()

    def display(self, msg: str):
        self.statements.append(msg)
        # print(msg)

    def visit_prog(self, prog: Prog, prefix=None):
        self.display(prefix + '主程序')
        for s in prog.statements:
            self.visit(s, prefix + blanks)

    def visit_variable_statement(self, var_statement: VariableStatement, prefix):
        self.display(prefix + '变量声明语句 ')
        self.visit(var_statement.variable_declare, prefix + blanks)

    def visit_variable_declare(self, var_declare: VariableDeclare, prefix):
        self.display(prefix + '声明变量 ' + var_declare.name)
        if var_declare.exp is not None:
            self.visit(var_declare.exp, prefix + blanks)
        else:
            self.display(prefix + blanks + ' 没有声明表达式.')

    def visit_parameter_list(self, param_list: ParameterList, prefix):
        self.display(prefix + '参数列表:')
        for p in param_list.params:
            self.visit(p, prefix + blanks)

    def visit_block(self, block: Block, prefix):
        self.display(prefix + '程序片段')
        for b in block.statements:
            self.visit(b, prefix + blanks)

    def visit_expression_statement(self, statement: ExpressionStatement, prefix):
        self.display(prefix + '表达式语句')
        return self.visit(statement.exp, prefix + blanks)

    def visit_empty_statement(self, statement: EmptyStatement, prefix):
        self.display(prefix + '空语句')

    def visit_binary(self, binary: Binary, prefix):
        self.display(prefix + '二元表达式 ' + binary.op.name)
        self.display(prefix + blanks + '二元表达式 左侧式子:')
        self.visit(binary.exp1, prefix + blanks * 2)
        self.display(prefix + blanks + '二元表达式 右侧式子:')
        self.visit(binary.exp2, prefix + blanks * 2)

    def visit_unary(self, unary: Unary, prefix):
        self.display(prefix + unary.op.name + '一元表达式')
        self.visit(unary.exp, prefix + blanks)

    def visit_integer_literal(self, literal: IntegerLiteral, prefix):
        self.display(prefix + '整数字面值: ' + str(literal.value))

    def visit_decimal_literal(self, literal: DecimalLiteral, prefix):
        self.display(prefix + '浮点数字面值: ' + str(literal.value))

    def visit_string_literal(self, literal: StringLiteral, prefix):
        self.display(prefix + '字符串字面值: ' + literal.value)

    def visit_array_literal(self, array: ArrayLiteral, prefix):
        self.display(prefix + '数组:' + str(array))
        i = 1
        for elem in array.exps:
            self.display(prefix + blanks + '数组第' + str(i) + '要素:')
            self.visit(elem, prefix + blanks * 2)
            i = i + 1

    def visit_map_literal(self, map_exp: MapLiteral, prefix):
        self.display(prefix + '字典:' + str(map_exp))
        i = 1
        for elem in map_exp.exps:
            self.display(prefix + blanks + '字典第' + str(i) + '要素Key:')
            self.visit(elem['key'], prefix + blanks * 2)
            self.display(prefix + blanks + '字典第' + str(i) + '要素value:')
            self.visit(elem['val'], prefix + blanks * 2)
            i = i + 1

    def visit_variable(self, variable: Variable, prefix):
        self.display(prefix + '变量:' + variable.name)

    def visit_function_call(self, func_call: FunctionCall, prefix):
        self.display(prefix + '函数调用:' + func_call.name)
        i = 1
        for param in func_call.params:
            self.display(prefix + blanks + '函数参数' + str(i))
            self.visit(param, prefix + blanks * 2)
            i = i + 1

    def visit_dot_exp(self, dot_exp: DotExp, prefix):
        self.display(prefix + '点运算')
        self.display(prefix + '  基础部分:')
        self.visit(dot_exp.base_exp, prefix + blanks * 2)
        self.display(prefix + '  指标部分:')
        self.visit(dot_exp.prop, prefix + blanks * 2)




demo1 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-10'
正回购资产_明细 = 资产(['债券'], ['久期', '持仓市值']) in 产品(['净值类'], ['总资产', '净资产', '杠杆率'])
正回购资产合计 = agg(正回购资产_明细, ['产品代码', '产品简称'], {'持仓市值': 'sum', '净资产': 'first', '杠杆率': 'first'})
持仓占比 = alias(正回购资产合计.持仓市值 / 正回购资产合计.净资产, '正回购占比')
监控规则 = {'触发超限': '正回购占比 > 0.44', '触发预警': '(正回购占比 <= 0.44) & (正回购占比 > 0.41)', '正常': '正回购占比 <= 0.41'}
'''

demo2 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-20'
资产指标库 = 指标('估值明细指标库', ['持仓市值'], {'port_tags': ['公募开放式', '封闭式'], 'asset_tags': ['债券', '债券回购']})
'''

demo3 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-20'
资产指标库 = 指标('估值明细指标库', ['持仓市值'], {'port_tags': ['公募开放式', '封闭式'], 'asset_tags': ['债券', '债券回购']})
最大持仓市值 = max(资产指标库, '持仓市值')
最小持仓市值 = min(资产指标库, '持仓市值')
平均持仓市值 = mean(资产指标库, '持仓市值')
持仓个数 = count(资产指标库, '持仓市值')
持仓中位数 = median(资产指标库, '持仓市值')
排序后资产指标库 = sort(资产指标库, ['持仓市值'], '降序')
'''

demo4 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-20'
资产指标库 = 指标('估值明细指标库', ['持仓市值'], {'port_tags': ['公募开放式', '封闭式'], 'asset_tags': ['债券', '债券回购']})
债券库 = filter(资产指标库, '债券 == "是"')
回购库 = filter(资产指标库, '债券回购 == "是"')
全部库 = concat(债券库, 回购库)
'''

demo5 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-20'
资产指标库 = 指标('估值明细指标库', ['持仓市值'], {'port_tags': ['公募开放式', '封闭式'], 'asset_tags': ['债券', '债券回购']})
排序后指标库 = sort(资产指标库, ['持仓市值'], '降序')
前十大持仓资产 = limit(排序后指标库, 10)
'''

demo6 = '''开始日期 = '2024-04-10'
结束日期 = '2024-04-20'
port_tags = ['公募开放式', '封闭式']
asset_tags = ['债券', '债券回购']
资产指标库 = 指标('估值明细指标库', ['持仓市值'], {'port_tags': port_tags, 'asset_tags': asset_tags})
排序后指标库 = sort(资产指标库, ['持仓市值'], '降序')
前十大持仓资产 = limit(排序后指标库, 10)
'''



# 内置金融算法库的调用和执行
import pandas as pd
import polars as pl
import muse.data_interface as t
from muse.utils.code_utils import run_script_in_venv, get_files_in_dir
# from muse.muse_config import PYTHON_PATH
import json
# from empyrical import stats
import numpy as np
from datetime import datetime, timedelta
from pathlib import Path
import os
import logging

def get_result_col(columns):
    score_list = list()
    for col in columns:
        score = 0
        if '-' in str(col):
            score += 1
        if '+' in str(col):
            score += 1
        if '*' in str(col):
            score += 1
        if '/' in str(col):
            score += 1
        score_list.append(score)

    result_col = None
    if len(score_list):
        max_value = max(score_list)  # 最大值
        if max_value > 0:
            max_index = score_list.index(max_value)  # 第一个最大值的索引
            result_col = columns[max_index]
    return result_col

def algo_handler(algo_name, params):
    logging.debug('调用算法函数: {}, 参数: {}, 调用结果: {}'.format(algo_name, params, 4))
    if algo_name == 'agg':
        df = params[0]
        group_keys = params[1]
        method_name = params[2]
        alias_name = None
        if len(params) > 3:
            alias_name = params[3]
        if method_name == 'single_ann_return':
            # method_name = stats.cum_returns_final
            method_name = np.sum
        return agg_algo(df, group_keys, method_name, alias_name)
    elif algo_name == '平台日':
        return platform_date()
    elif algo_name == '近N天':
        current_date = params[0]
        days = params[1]
        return n_days(current_date, days)
    elif algo_name == '近N月':
        current_date = params[0]
        months = params[1]
        return n_months(current_date, months)
    elif algo_name == 'sum':
        df = params[0]
        col_name = params[1]
        return sum_algo(df, col_name)
    elif algo_name == 'mean':
        df = params[0]
        col_name = params[1]
        return mean_algo(df, col_name)
    elif algo_name == 'median':
        df = params[0]
        col_name = params[1]
        return median_algo(df, col_name)
    elif algo_name == 'min':
        df = params[0]
        col_name = params[1]
        return min_algo(df, col_name)
    elif algo_name == 'max':
        df = params[0]
        col_name = params[1]
        return max_algo(df, col_name)
    elif algo_name == 'count':
        df = params[0]
        col_name = params[1]
        return count_algo(df, col_name)
    elif algo_name == 'sort':
        df = params[0]
        sort_by = params[1]
        is_asc = None
        if len(params) > 2:
            is_asc = params[2]
        return sort_algo(df, sort_by, is_asc)
    elif algo_name == 'limit':
        df = params[0]
        limit_no = params[1]
        return limit_algo(df, limit_no)
    elif algo_name == 'alias':
        df = params[0]
        alias_name = params[1]
        if len(params) > 2:
            col_to_change = params[2]
        else:
            col_to_change = None
        return alias_algo(df, alias_name, col_to_change)
    elif algo_name == 'merge':
        df1 = params[0]
        df2 = params[1]
        on_cols = params[2]
        how = params[3]
        return merge_algo(df1, df2, on_cols, how)
    elif algo_name == 'concat':
        df1 = params[0]
        df2 = params[1]
        return concat_algo(df1, df2)
    elif algo_name == 'select':
        df = params[0]
        cols = params[1]
        return select_algo(df, cols)
    # elif algo_name == '调用':
    #     file_locate = params[0]
    #     params = params[1]
    #     return invoke_py_algo(file_locate, params)
    return 0

def concat_algo(df1, df2):
    if df1.is_empty() and not df2.is_empty():
        return df2
    elif not df1.is_empty and df2.is_empty():
        return df1
    elif df1.is_empty and df2.is_empty():
        return pl.DataFrame()

    df = pl.concat([df1, df2], how='vertical')
    return df


def select_algo(df, cols):
    if df.is_empty():
        return df
    else:
        existing_columns = [col for col in cols if col in df.columns]
        selected_df = df.select(existing_columns)
        return selected_df

def select_algo(df, cols):
    if df.is_empty():
        return df
    else:
        existing_columns = [col for col in cols if col in df.columns]
        selected_df = df.select(existing_columns)
        return selected_df

def platform_date():
    current_date = datetime.now()
    return current_date.strftime('%Y-%m-%d')


def n_days(current_date, days):
    current_d = datetime.strptime(current_date, '%Y-%m-%d')
    previous_date = current_d - timedelta(days=days)
    return previous_date.strftime('%Y-%m-%d')


def n_months(current_date, months):
    current_d = datetime.strptime(current_date, '%Y-%m-%d')
    previous_date = current_d - timedelta(days=months * 30)
    return previous_date.strftime('%Y-%m-%d')


def agg_algo(df: pl.DataFrame, group_keys: list, algos:str, alias_name=None):
    # 假设没有任何额外指标提供，资产代码又在dataframe里，用资产代码作为该字段
    # 支持字典
    if df is None or df.is_empty():
        return pl.DataFrame()
    if isinstance(algos, dict):
        oper_list = []
        for ind, algo_name in algos.items():
            alias = ind
            if alias_name and isinstance(alias_name, dict):
                if ind in alias_name.keys():
                    alias = alias_name[ind]
            if ind in df.columns:
                if algo_name == 'sum':
                    oper_list.append(pl.col(ind).fill_null(0).sum().alias(alias))
                elif algo_name == 'mean' or algo_name == 'avg' or algo_name == 'average':
                    oper_list.append(pl.col(ind).fill_null(0).mean().alias(alias))
                elif algo_name == 'count' or algo_name == 'len':
                    oper_list.append(pl.col(ind).fill_null(0).len().alias(alias))
                elif algo_name == 'first':
                    oper_list.append(pl.col(ind).fill_null(0).first().alias(alias))
                elif algo_name == 'last':
                    oper_list.append(pl.col(ind).fill_null(0).last().alias(alias))
                elif algo_name == 'max':
                    oper_list.append(pl.col(ind).fill_null(0).max().alias(alias))
                elif algo_name == 'min':
                    oper_list.append(pl.col(ind).fill_null(0).min().alias(alias))
                elif algo_name == 'concat' or algo_name == 'cat':
                    oper_list.append(pl.col(ind).str.concat(",").alias(alias))
        lz = (
            df.lazy().group_by(group_keys).agg(oper_list)
        )
    else:
        if alias_name is not None and isinstance(alias_name, dict) and len(alias_name.keys()) > 0:
            ind = list(alias_name.keys())[0]
            alias = alias_name[ind]

        else:
            ind = t.find_prop(list(df.columns))
            if ind is None:
                ind = list(set(df.columns) - set(group_keys))[0]
            alias = ind
        if algos == 'sum':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).sum().alias(alias))
            )
        elif algos == 'mean' or algos == 'avg' or algos == 'average':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).mean().alias(alias))
            )
        elif algos == 'count' or algos == 'len':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).len().alias(alias))
            )
        elif algos == 'first':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).first().alias(alias))
            )
        elif algos == 'last':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).last().alias(alias))
            )
        elif algos == 'max':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).max().alias(alias))
            )
        elif algos == 'min':
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).min().alias(alias))
            )
        else:
            lz = (
                df.lazy().group_by(group_keys).agg(pl.col(ind).fill_null(0).sum().alias(alias))
            )
    df = lz.collect()
    return df


def sum_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().sum()


def mean_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().mean()


def max_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().max()


def min_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().min()


def median_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().median()


def count_algo(df: pl.DataFrame, col_name: str):
    if df.is_empty():
        return 0
    return df.select(col_name).to_series().len()


def sort_algo(df: pl.DataFrame, sort_bys, is_asc=None):
    if df.is_empty():
        return df
    if is_asc is None or is_asc == '升序':
        lz = (df.lazy().sort(by=sort_bys, descending=False))
        return lz.collect()
    else:
        lz = (df.lazy().sort(by=sort_bys, descending=True))
        return lz.collect()


def limit_algo(df: pl.DataFrame, limit_no):
    if df.is_empty():
        return df
    return df.head(limit_no)


def alias_algo(df: pl.DataFrame, alias_name: str, col_to_change):
    if not df.is_empty():
        cols = list(df.columns)
        if col_to_change is None:
            col_to_change = get_result_col(cols)
        df = df.rename({col_to_change: alias_name})
        return df
    else:
        return df


def merge_algo(df1: pl.DataFrame, df2: pl.DataFrame, on_cols: list, how='left'):
    if df1.is_empty() and not df2.is_empty():
        return df2
    elif not df1.is_empty() and df2.is_empty():
        return df1
    else:
        return df1.join(df2, on=on_cols, how=how)

# def invoke_py_algo(file_locate: str, params: dict):
#     params_str = json.dumps(params, ensure_ascii=False)
#     python_result = run_script_in_venv(PYTHON_PATH, file_locate, params_str)
#     if python_result['success']:
#         result = dict()
#         parent_dir = str(os.path.dirname(file_locate))
#         if python_result['success']:
#             new_files = get_files_in_dir(parent_dir)
#             for file_addr in new_files:
#                 file_path = Path(file_addr)
#                 ext = file_path.suffix
#                 if ext.lower() == '.csv':
#                     base_name = os.path.basename(file_addr)
#                     file_name = os.path.splitext(base_name)[0]
#                     df = pl.read_csv(file_addr)
#                     result[file_name] = df
#         return result
#     else:
#         raise Exception(python_result['stderr'])





import requests

class JxHttpClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def get(self, endpoint: str, params: dict = None):
        """发送 GET 请求"""
        url = f"{self.base_url}{endpoint}"
        try:
            response = requests.get(url, params=params)
            response.raise_for_status()  # 抛出异常以处理非 2xx 状态码
            return response.json()  # 返回 JSON 响应内容
        except requests.HTTPError as e:
            print(f"HTTP错误: {e}")
        except ValueError:
            print("返回的内容无法解析为 JSON")
        except Exception as e:
            print(f"发生错误: {e}")

    def post(self, endpoint: str, data: dict):
        """发送 POST 请求"""
        url = f"{self.base_url}{endpoint}"
        headers = {
            "Content-Type": "application/json;charset=UTF-8"
        }
        try:
            response = requests.post(url, json=data, headers=headers)
            response.raise_for_status()  # 抛出异常以处理非 2xx 状态码
            return response.json()  # 返回 JSON 响应内容
        except requests.HTTPError as e:
            print(f"HTTP错误: {e}")
        except ValueError:
            print("返回的内容无法解析为 JSON")
        except Exception as e:
            print(f"发生错误: {e}")

    def put(self, endpoint: str, data: dict):
        """发送 PUT 请求"""
        url = f"{self.base_url}{endpoint}"
        headers = {
            "Content-Type": "application/json;charset=UTF-8"
        }
        try:
            response = requests.put(url, json=data, headers=headers)
            response.raise_for_status()  # 抛出异常以处理非 2xx 状态码
            return response.json()  # 返回 JSON 响应内容
        except requests.HTTPError as e:
            print(f"HTTP错误: {e}")
        except ValueError:
            print("返回的内容无法解析为 JSON")
        except Exception as e:
            print(f"发生错误: {e}")

    def delete(self, endpoint: str):
        """发送 DELETE 请求"""
        url = f"{self.base_url}{endpoint}"
        try:
            response = requests.delete(url)
            response.raise_for_status()  # 抛出异常以处理非 2xx 状态码
            return response.json() if response.content else {}  # 如果没有内容，返回空字典
        except requests.HTTPError as e:
            print(f"HTTP错误: {e}")
        except ValueError:
            print("返回的内容无法解析为 JSON")
        except Exception as e:
            print(f"发生错误: {e}")


import polars as pl
from datetime import datetime
def filter_dates(pdf: pl.DataFrame, start_date, end_date):
    if '日期' in pdf.columns:
        pdf = pdf.with_columns(
            pl.col("日期").str.to_date("%Y-%m-%d")
        )
        if start_date and end_date:
            start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
            end_date = datetime.strptime(end_date, "%Y-%m-%d").date()
            return pdf.filter(pl.col("日期").is_between(start_date, end_date, closed="both"))
    return pdf


# 根据产品代码或者资产代码过滤
def filter_ports_sec(pdf: pl.DataFrame, port_ids: list = [], sec_ids: list = [], start_date=None, end_date=None):
    pdf = filter_dates(pdf, start_date, end_date)
    if (port_ids is not None) and (len(port_ids) > 0) and (
            '查询所有组合' not in port_ids) and '产品代码' in pdf.columns:
        pdf = pdf.filter(pl.col("产品代码").is_in(port_ids))
    if sec_ids is not None and len(sec_ids) > 0 and '资产代码' in pdf.columns:
        pdf = pdf.filter(pl.col("资产代码").is_in(sec_ids))
    return pdf


def filter_ports(pdf: pl.DataFrame, port_ids: list = [], start_date=None, end_date=None):
    pdf = filter_dates(pdf, start_date, end_date)
    if (port_ids is not None) and (len(port_ids) > 0) and ('查询所有组合' not in port_ids):
        pdf = pdf.filter(pl.col("产品代码").is_in(port_ids))
    return pdf

def filter_hld_data(df: pl.dataframe, penetrate: str):
    ctlx = []
    if penetrate is None:
        ctlx = ['00', '01', '02']
    elif penetrate == '不穿透':
        ctlx = ['00', '01', '02']
    elif penetrate == '全穿透':
        ctlx = ['00', '11', '12']
    elif penetrate == '自主管理穿透':
        ctlx = ['00', '11', '02']
    elif penetrate == '委外投资穿透':
        ctlx = ['00', '01', '12']
    df = df.filter(pl.col("穿透类型").is_in(ctlx))
    return df

def convert_decimal_to_float(df: pl.DataFrame) -> pl.DataFrame:
    exprs = []
    for col_name, dtype in df.schema.items():
        if str(dtype).startswith("Decimal"):
            exprs.append(pl.col(col_name).cast(pl.Float64).alias(col_name))
        # 其他类型（如 Struct/List）可在此扩展
    return df.with_columns(exprs) if exprs else df

def get_all_unique_fields(cols: list, basic_cols: list) -> list:
    """返回 cols 和 basic_cols 的并集（去重后的所有字段）

    Args:
        cols: 额外字段列表
        basic_cols: 基础字段列表

    Returns:
        去重后的所有字段列表（顺序不保证）
    """
    return list(set(cols) | set(basic_cols))


from muse.data_source.base import DataSourceInterface
from muse.data_source.ds_utils import filter_ports, filter_ports_sec, convert_decimal_to_float, get_all_unique_fields
import polars as pl
import os
from muse.utils.jx_http_request import JxHttpClient
from collections import OrderedDict
from muse.muse_config import DATA_PATH, MID_BASE_URL, ASSET_TAG_GROUP, PORT_TAG_GROUP, REPOSITORY_GROUP

class DataSourceApi(DataSourceInterface):
    def get_datasource_name(self):
        return f'指标库数据源 - 地址: {MID_BASE_URL}'

    def get_basic_inds(self):
        return ['资产代码', '资产名称', '资产标签', '产品代码', '产品简称', '产品标签', '主体代码', '主体名称', '主体标签', '日期']

    def get_port_tags(self):
        # tags = pl.read_excel(self.data_locate + '产品标签列.xlsx')
        jxClient = JxHttpClient(MID_BASE_URL)
        port_targs_params = {
            "classCode": 80,
            "grpName": "限额指标产品标签组",
            "curUser": {
                "realName": "系统管理员",
                "userName": "padmin",
                "orgCode": "BM0008"
            }
        }
        rst = jxClient.post("/labelSystem/getLabelCloud", data=port_targs_params)
        rst_list = [item['typeName'] for item in rst if item['count'] > 0]
        return rst_list

    def get_port_inds(self):
        inds = pl.read_excel(DATA_PATH + '/产品指标列.xlsx')
        return inds.columns

    def get_asset_tags(self):
        jxClient = JxHttpClient(MID_BASE_URL)
        port_targs_params = {
            "classCode": 81,
            "grpName": ASSET_TAG_GROUP,
            "curUser": {
                "realName": "系统管理员",
                "userName": "padmin",
                "orgCode": "BM0008"
            }
        }
        rst = jxClient.post("/labelSystem/getLabelCloud", data=port_targs_params)
        rst_list = [item['typeName'] for item in rst if item['count'] > 0]
        return rst_list

    def get_asset_inds(self):
        inds = pl.read_excel(DATA_PATH + '/资产指标列.xlsx')
        return inds.columns

    def get_data(self, run_method: str, data_subject: str, sec_ids: list = [], port_ids: list = [], tags: list = [],
                 inds: list = [], start_date=None, end_date=None, penetrate=None):
        if penetrate:
            penetrate = '全穿透'
        else:
            penetrate = '不穿透'
        pdf = None

        if data_subject == '持仓指标':
            asset_df = pl.read_excel(DATA_PATH + '/资产指标汇总.xlsx')
            asset_tags = pl.read_excel(DATA_PATH + '/资产标签.xlsx')
            asset_df = asset_df.join(asset_tags, on=['资产代码'], how='left')
            pdf = filter_ports_sec(pdf=asset_df, port_ids=port_ids, sec_ids=sec_ids, start_date=start_date,
                                   end_date=end_date)
            # 获得产品代码
            port_ids = pdf.select("产品代码").unique().to_series().to_list()
            ports = pl.read_excel(DATA_PATH + '/产品指标汇总.xlsx')
            port_tags = pl.read_excel(DATA_PATH + '/产品标签.xlsx')
            ports = ports.join(port_tags, on=['产品代码'], how='left')
            ports = filter_ports(pdf=ports, port_ids=port_ids, start_date=start_date, end_date=end_date)
            # 合并资产和产品
            pdf = pdf.join(ports, on=['产品代码', '日期'], how='left')
            pdf = convert_decimal_to_float(pdf)
        elif data_subject == '产品指标':
            ports = pl.read_excel(self.data_locate + '/产品指标汇总.xlsx')
            port_tags = pl.read_excel(self.data_locate + '/产品标签.xlsx')
            ports = ports.join(port_tags, on=['产品代码'], how='left')
            pdf = filter_ports_sec(pdf=ports, port_ids=port_ids, start_date=start_date, end_date=end_date)

        merged_cols = tags + inds
        if len(merged_cols) > 0:
            all_cols = get_all_unique_fields(merged_cols, self.get_basic_inds())
            existing_cols = [c for c in all_cols if c in pdf.columns]
            pdf = pdf.select(existing_cols)
        return pdf

    # 根据标签获取id列表
    def get_ids_by_tags(self, tags: list = [], tag_grps = []):
        tag_grps_list = [PORT_TAG_GROUP, ASSET_TAG_GROUP]
        if tag_grps:
            tag_grps_list = tag_grps

        api_params = {
            "grpNames": tag_grps_list,
            "labelNames": tags
        }
        jxClient = JxHttpClient(MID_BASE_URL)
        rst = jxClient.post("/labelSystem/getIdsByLabelName", data=api_params)
        return rst

    def get_repository_inds(self, run_method: str, repository: str, ind_list: list, params: dict):
        # 基础API参数
        api_params = {
            "resource": repository,
            "paramMaps": {"SYS_ID": ""},
            "ind_name_list": ind_list,
            "pageSize": 1000000,
            "pageNum": 1,
            "curUser": {"userName": "test001"}
        }

        # 处理日期参数, 如果传入了小写的，则转成大写
        date_mapping = {
            'start_date': 'START_DATE',
            'end_date': ['END_DATE', 'DATA_DATE']
        }

        for src_key, dest_keys in date_mapping.items():
            if src_key in params:
                value = params.pop(src_key)
                if value:
                    # 统一将目标键转换为列表处理
                    target_keys = dest_keys if isinstance(dest_keys, list) else [dest_keys]
                    for target_key in target_keys:
                        params[target_key] = value

        ### ---------- 处理 port_ids ---------- 开始
        port_ids = []
        if 'port_ids' in params:
            port_ids_param = params.pop('port_ids', '')
            # 处理逗号分隔的字符串，转换为列表
            if isinstance(port_ids_param, str):
                port_ids = [item.strip() for item in port_ids_param.split(',') if item.strip()]
            elif isinstance(port_ids_param, list):
                port_ids = port_ids_param

        # 再处理 port_tags
        if 'PORT_TAGS' in params:
            port_tags = params.pop('PORT_TAGS')  # 使用 pop 移除参数
            # 处理逗号分隔的字符串，转换为列表
            if isinstance(port_tags, str):
                port_tags = [tag.strip() for tag in port_tags.split(',') if tag.strip()]

            if isinstance(port_tags, list) and port_tags:
                port_tags_ids = self.get_ids_by_tags(tags=port_tags)
                # 合并并去重
                if port_tags_ids:
                    port_ids = list(set(port_ids + port_tags_ids))

        if port_ids:
            params['PORT_IDS'] = ",".join(port_ids)
        ### ---------- 处理 port_ids ---------- 结束

        ### ---------- 处理 sec_ids ---------- 开始
        sec_ids = []
        if 'SEC_IDS' in params:
            sec_ids_param = params.pop('sec_ids', '')
            # 处理逗号分隔的字符串，转换为列表
            if isinstance(sec_ids_param, str):
                sec_ids = [item.strip() for item in sec_ids_param.split(',') if item.strip()]
            elif isinstance(sec_ids_param, list):
                sec_ids = sec_ids_param

        if 'SEC_TAGS' in params:
            sec_tags = params.pop('SEC_TAGS')  # 使用 pop 移除参数
            # 处理逗号分隔的字符串，转换为列表
            if isinstance(sec_tags, str):
                sec_tags = [tag.strip() for tag in sec_tags.split(',') if tag.strip()]

            if isinstance(sec_tags, list) and sec_tags:
                sec_tags_ids = self.get_ids_by_tags(tags=sec_tags)
                if sec_tags_ids:
                    sec_ids = list(set(sec_ids + sec_tags_ids))

        if sec_ids:
            params['SEC_IDS'] = ",".join(sec_ids)
        ### ---------- 处理 sec_ids ---------- 结束

        api_params["paramMaps"] = {**api_params["paramMaps"], **params}
        jxClient = JxHttpClient(MID_BASE_URL)
        rst = jxClient.post("/xlsApiPlugin/museLoadData", data=api_params)
        # 提取列名
        column_names = [header['ind_name'] for header in rst['headers']]

        df = create_dataframe_safely(rst['datas'], column_names)
        return df

    def get_repository_metadata(self, rep_list: list):
        rep_locate = f'{DATA_PATH}/指标库/{rep_list[0]}.xlsx'
        if not os.path.exists(rep_locate):
            return {}
        # 选择合适的指标
        metadata = dict()
        metadata['指标库名称'] = rep_list[0]

        df = pl.read_excel(rep_locate)
        # 指标描述需要额外获取，描述应包含所有枚举值信息; 指标类型信息
        ind_list = [{'指标名称': ind, '指标描述': ind, '指标类型': 'string'} for ind in df.columns]
        metadata['指标列表'] = ind_list
        # 每个指标库都应该一样的部分
        params = {
                    'port_ids': {'参数描述': '产品代码列表', '参数类型': 'list'},
                    'port_tags': {'参数描述': '产品标签列表', '参数类型': 'list'},
                    'asset_tags': {'参数描述': '资产标签列表', '参数类型':'list'},
                    'start_date': {'参数描述': '查询开始日期', '参数类型': 'date'},
                    'end_date': {'参数描述': '查询结束日期', '参数类型': 'data'}
                  }
        # 每个指标库不一样的部分放到下面填补
        metadata['参数列表'] = params
        return metadata


    def get_all_repository_metadata(self):
        jxClient = JxHttpClient(MID_BASE_URL)
        original_data_list = jxClient.get(f"/museapi/loadMuseInds?topics={REPOSITORY_GROUP}")

        if not original_data_list:
            return []  # 如果没有数据，返回空列表

        transformed_data_list = []  # 存储所有转换后的指标库

        # 循环处理每一个指标库
        for original_data in original_data_list:
            transformed_data = OrderedDict()

            # 转换基本信息
            transformed_data['指标库名称'] = original_data['apiName']
            transformed_data['指标库描述'] = original_data['apiDesc']

            # 转换参数列表
            transformed_data['参数列表'] = {}
            for param_key, param_value in original_data['apiParamMap'].items():
                transformed_param = {
                    '参数描述': param_value['apiParamName'],
                    '参数类型': param_value['apiParamType'],
                    '参数必填': '是' if param_value['apiParamRequired'] else '否',
                    '参数默认值': param_value['defaultValue'],
                    '参数字典值(key: 参数含义 Value: 调用时应填值)': param_value['apiParamDict'] if param_value[
                        'apiParamDict'] else '无'
                }
                transformed_data['参数列表'][param_key] = transformed_param

            if 'PORT_IDS' in transformed_data['参数列表']:
                PORT_TAGS_PARAMS = {
                    '参数描述': '逗号分隔的产品标签列表',
                    '参数类型': 'String'
                }
                transformed_data['参数列表']['PORT_TAGS'] = PORT_TAGS_PARAMS

            if 'SEC_IDS' in transformed_data['参数列表']:
                SEC_TAGS_PARAMS = {
                    '参数描述': '逗号分隔的资产标签列表',
                    '参数类型': 'String'
                }
                transformed_data['参数列表']['SEC_TAGS'] = SEC_TAGS_PARAMS

            # 转换指标列表
            transformed_data['指标列表'] = []
            for indicator in original_data['indList']:
                ind_name = indicator['indName']
                ind_desc = indicator['indDesc']
                if ind_name == '产品代码':
                    ind_desc = '理财产品的系统唯一标识码，指标库重要关联指标'
                elif ind_name == '资产代码' or ind_name == '证券代码':
                    ind_desc = '理财产品投资的金融资产的系统唯一标识码, 指标库重要关联指标'
                elif '日期' in ind_name:
                    ind_desc = '表示数据统计日期, 指标库重要关联指标。数据表示为YYYY-mm-dd的日期字符串'

                transformed_indicator = {
                    '指标名称': ind_name,
                    '指标描述': ind_desc
                }
                transformed_data['指标列表'].append(transformed_indicator)

            transformed_data_list.append(transformed_data)

        return transformed_data_list  # 返回转换后的列表

    def get_repository_params(self, rep_list: list):
        jxClient = JxHttpClient(MID_BASE_URL)
        params = jxClient.get(f"/museapi/getApiParamListByApiCodes?apiCodes={','.join(rep_list)}")
        return params


def create_dataframe_safely(data, schema):
    """安全创建DataFrame，处理混合数据类型问题"""
    # 方法1：先尝试正常创建
    try:
        return pl.DataFrame(data, schema=schema, orient='row', infer_schema_length=10000)
    except Exception as e:
        print(f"方法1失败: {e}")

    # 方法2：逐列构建
    data_dict = {}
    for col_idx, col_name in enumerate(schema):
        col_data = []
        for row in data:
            if col_idx < len(row):
                value = row[col_idx]
                # 统一数值类型
                if isinstance(value, (int, float)):
                    col_data.append(float(value))
                else:
                    col_data.append(value)
            else:
                col_data.append(None)
        data_dict[col_name] = col_data

    return pl.DataFrame(data_dict)

if __name__ == '__main__':
    ds = DataSourceApi()
    params={
        "START_DATE": "2023-03-01",
        "END_DATE": "2023-04-05",
        "PORT_TAGS": "现金管理类产品"
    }
    print(ds.get_repository_inds(run_method='', repository='产品估值（时序）', ind_list=['产品代码', '统计日期', '总资产', '净资产'], params=params))
    print(ds.get_repository_inds(run_method='', repository='产品期间收益时点指标库', ind_list=[], params=params))
    print(ds.get_all_repository_metadata())
    print(ds.get_asset_tags())
    print(ds.get_asset_inds())
    print(ds.get_port_tags())
    print(ds.get_port_inds())


from muse.data_source.demo_ds import DemoDataSource
from muse.data_source.data_source_api import DataSourceApi
from muse.data_source.base import DataSourceInterface

class _DataSourceManager:
    _data_source = DataSourceApi()
    def init_ds(self, ds):
        if isinstance(ds, DataSourceInterface):
            self._data_source = ds
            ds_name = ds.get_datasource_name()
            print(f'加载数据源: "{ds_name}" ...... 成功!')
    def get_ds(self):
        ds_name = self._data_source.get_datasource_name()
        # print(f'正在使用数据源: "{ds_name}" ......')
        return self._data_source

ds_manager = _DataSourceManager()